#! /usr/bin/env bash

# Arguments: inputBatchSaveFile outputBatchSaveFile
#   bash @shellscript @RServerScript @rcall @result stdout=@stout stderr=@sterr;


# Set restrictive umask for duration of script
# This value prevents any other users from reading
# or writing
umask "077"

#if [ $# != 4 ]; then
if [ $# -lt 4 ]; then
  echo $0: expecting 4+ arguments, got $#: $* 1>&2
  exit 1
fi

# tmp=/tmp # FIXME: allow this to change eg for sites with main tmp dir elsewhere
# tmp=/scratch/local # FIXME: allow this to change eg for sites with main tmp dir elsewhere
tmp=${SWIFTR_TMP:-${TMP:-/tmp}}

RServerScript=$1
callFile=$2
resultFile=$3
shift; shift; shift;

if [ $1 != "--imports" ];
then
    echo $0: expected --imports to be provided
fi
shift;
# now $@ contains imports
imports=$@



# Find our bin dir (to use for running utility scripts)

SWIFTRBIN=$(cd $(dirname $0); pwd)

# This script expects to be called (typically from Swift worker.pl)
# with the following env vars set:
#  SWIFT_WORKER_PID - process ID of the worker.pl that launched this script
#  SWIFT_JOB_SLOT   - slot ID of the worker job that launched this script (0..maxWorkers-1)

if [ _$SWIFT_WORKER_PID = _ ]; then
  echo $0: SWIFT_WORKER_PID is not set. SWIFT_JOB_SLOT=$SWIFT_JOB_SLOT ARGS=$* 1>&2
  exit 1
fi
if [ _$SWIFT_JOB_SLOT = _ ]; then
  echo $0: SWIFT_JOB_SLOT is not set. SWIFT_WORKER_PID=$SWIFT_WORKER_PID ARGS=$* 1>&2
  exit 1
fi

TRAPS="EXIT 1 2 3 15"

function onexit() { # FIXME: move this logic into worker.pl, or try to use R timeout option if it works on fifos - need to test.
  trap - $TRAPS
  echo "EvalRBatchPersistent.sh: onexit trap: SIGNAL=$1 RPIDS=$RPIDS"
  if [ "_$RPIDS" != _ ]; then
    kill $RPIDS >& /dev/null
  fi
}


trap "onexit 1" 1
trap "onexit 2" 2
trap "onexit 3" 3
trap "onexit 15" 15
trap "onexit EXIT" EXIT


function idletimer {
  cd $1
  while true; do
    touch idle
    sleep 60
    if [ ! -f lastwrite -o \( idle -nt lastwrite \) ]; then
      # FIXME: should work, but does not: echo quit x x >toR.fifo # fifo problem?
      RPID=$(cat R.pid)
      echo killing idle R process $RPID
      kill $RPID
      exit 0
    fi
  done
}

# Ensure that the dir for this slot exists. 

BASEDIR=$tmp/$(id -nu)/SwiftR/Rworkers
WORKERDIR=$BASEDIR/worker.$SWIFT_WORKER_PID
SLOTDIR=$WORKERDIR/${SWIFT_JOB_SLOT}

mkdir -p $WORKERDIR

RPIDS=

TIMEOUT=5

function start_timeout {
    ppid=$$
    trap "timeout_handler" SIGHUP
    (
    sleep ${TIMEOUT}s
    kill -1 $ppid  &> /dev/null #SIGHUP 
    if [ "$1" != "" ]; then
        kill -1 $1  &> /dev/null #SIGHUP 
    fi
    ) &
}

function stop_timeout {
    #DEBUG
    trap "" SIGHUP
}

function timeout_handler {
    echo 'Timed out waiting to contact R process'
    echo 'R log follows:'
    echo '====================='
    cat $SLOTDIR/R.log
    exit 1
}



# Try to make slotdir.
# If the mkdir succeeds, this is the first request to the slot,
# so we create a new R server and send the current request to it;
# else we send the current request to an existing R server.

mkdir $SLOTDIR >& /dev/null
if [ $? = 0 ]; then
  mkfifo $SLOTDIR/toR.fifo
  mkfifo $SLOTDIR/fromR.fifo
  chmod +x $RServerScript
  echo "$0: INFO: Launching $RServerScript $SLOTDIR"
  #echo PATH: $PATH > $SLOTDIR/R.log
  #which Rscript >> $SLOTDIR/R.log
  
  echo LD_LIBRARYPATH= $LD_LIBRARY_PATH
  PATH=.:$PATH
  echo PATH= $PATH
  echo R_LIBS_USER= $R_LIBS_USER
  # Use R instead of Rscript due to Rscript issues on some
  # platforms (e.g. cray xt)
  R --slave --no-restore --file=$RServerScript \
                --args $SLOTDIR >> $SLOTDIR/R.log 2>&1 &   # launch R server
  #$RServerScript $SLOTDIR >> $SLOTDIR/R.log 2>&1 &   # launch R server
  # idletimer $SLOTDIR </dev/null >/dev/null 2>&1 & # R saves pid in R.pid for idletimer to kill it
  echo "$0: INFO: Launched $RServerScript $SLOTDIR Rscript"
else
  # wait to make sure fifo exists
  # fromR fifo is created last, so wait for that one
  while true; do 
    if [ -p $SLOTDIR/fromR.fifo ]; then
      break
    fi
  done
fi

# Ready to talk to the server: send request and read response

start_timeout
while true; do
  mkdir $SLOTDIR/mutex
  if [ $? != 0 ]; then
    sleep 1;
  else
    break;
  fi
done
stop_timeout
echo DB: Obtained $SLOTDIR/mutex

absimports=
for im in $imports; do
    absimports="$absimports $(pwd)/$im"
done
echo imports: $imports
echo absimports: $absimports
echo run $(pwd)/$callFile $(pwd)/$resultFile $absimports > $SLOTDIR/toR.fifo &
echopid=$!
start_timeout $echopid

if wait $echopid
then
    touch $SLOTDIR/lastwrite

    echo DB: Sent request
    stop_timeout # started up ok
    echo dummy stderr response 1>&2 # FIXME - testing if this is the provider staging problem (not xfering zero len stderr)

    res=$(cat < $SLOTDIR/fromR.fifo)
    echo DB: Got response: $res

    # Test if R server reported an error
    if echo "$res" | grep -q '^ERROR:'
    then
        exit 1
    fi

    rmdir $SLOTDIR/mutex

    echo DB: Freed $SLOTDIR/mutex
else
    stop_timeout
    echo "ERROR: Could not write to fifo ok"
    rmdir $SLOTDIR/mutex

    exit 1
fi

# Fixme: how to get exceptions and stdout/stderr text from R server ???
